#!/usr/bin/awk -f
#
# zpsoli 1.0 - portable Solitaire/Patience/Klondike card game in awk.
# Copyright (C) 2021 Russell Marks.
# Based on my 2002 Minigame Competition entry for the Sinclair Spectrum.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


# This is probably halfway between a port and a rewrite. The display
# code was redone to work line-by-line, for example. And, well, the
# original was in Z80 assembly, so that's quite a change too. :-)
#
# Why do an awk version? I wanted to do something fun in awk that was
# more than the usual random tiny program stuffed in a shell script. I
# also wanted to do it in "old awk" so hopefully it'd work on every
# version of awk... it's debatable whether I managed that properly.
# As-is this is actually a new-awk-only version using srand/rand - but
# the Makefile tweaks it to make an old-awk version too, and that can
# run on V7 awk as well as newer ones.
#
# This has mostly been tested on gawk 4/5 and (in SIMH) V7 awk, but
# it's also seemed to work on various other awks I've tried of various
# ages and origins (Minix 2's awk is one exception). Note that the
# game will probably be patience-testingly slow on most computers
# dating from before, say, 1990.


# Known bugs:
#
# Aces can be put on the main piles - this was probably the case in
# soliyawn/zcsoli as well. Harmless enough, but ideally that should be
# prohibited.


# Feature-testing notes:
#
# (Because I'd rather try stuff than rely purely on documentation...)
#
# gawk = gawk 4.2.1,
# mawk = mawk 1.3.3,
# bwkawk = BWK awk (as Debian's original-awk version 2012-12-20-6),
# bbawk = BusyBox 1.30.1 awk,
# s4nawk = SunOS 4.1.1 nawk (sun3),
# s4awk = as above but awk,
# v7awk = V7 awk (PDP-11).
#
# The gawk/mawk/BWK-awk/busybox versions used were from Debian Buster
# (well, Raspbian in my case). No special reason to test SunOS 4, it's
# only included because it's the first Unix I used. :-) It also has
# the "advantage" of having both an old and new awk available.
#
# feature	gawk	mawk	bwkawk	bbawk	s4nawk	s4awk	v7awk
#
# array[x":"y]	yes	yes	yes	yes	yes	yes	yes
# rand/srand	yes	yes	yes	yes	yes	no	no
# functions	yes	yes	yes	yes	yes	no	no
# system	yes	yes	yes	yes	yes	no	no
# "\033"	yes	yes	yes	yes	yes	no	no
# array[x,y]	yes	yes	yes	yes	yes	no	no
# myflag {	yes	yes	yes	yes	no	no	no
# array[x][y]	yes	no	no	no	no	no	no
#
# ("functions" above means user-defined functions. And it seems like
# foo[x "@" y] may be more usual, but I just preferred ":".)
#
# So V7 is clearly the lowest common denominator (unsurprisingly) and
# the obvious target. If you ever think "but why didn't he just...",
# see above. :-) Other things V7 awk lacks which impact the code
# include getline, and assignment to $0.


# Original soliyawn/zcsoli card encoding was one byte:
#
#;76543210
#;xxy0zzzz
#;
#;xx = suit, 0-3 respectively hearts, diamonds, spades, clubs
#;     (so top bit doubles up as red/black bit, 1=black).
#;0 = unused
#;y = set if card is face down
#;    (not applied for cards still in the deck, the top card
#;     there is shown face down `by hand' when appropriate)
#;zzzz = value, 1-13 respectively A, 2-10, J, Q, K.
#
# This version uses a three/four-char string for xyz, and "" as a
# no-more-cards-in-pile marker. So e.g. "001" is face-up ace of
# hearts, "3113" is face-down king of clubs. This means we can use
# substr() to get at each field, but still only need one array entry
# per card.
#
# [The downside is that if we somehow treat the string as a number,
# everything would break as leading zeroes would disappear. Luckily I
# seem to have avoided doing that, but were I to redo this I would
# use 1-4 for the suit digit.]
#
# The piles are in a single pseudo-two-dimensional array, used as
# piles[x ":" y] where x is pile number (0..13) and y is card number
# in pile (0..51). Pile numbers are just the same as earlier versions:
#
#;0	deck (in bottom-to-top order)
#;1	turned-over pile next to deck (ditto)
#;2-8	the main piles
#;9-12	the ace piles
#
# The ace "piles" are really just card slots, only storing the
# most-recently-placed card.
#
# (And to make things a bit faster/simpler, plast[0..13] holds the
# index of the last card on each pile, -1 for none.)
#
# Presumably maths on the array indicies is best avoided, since they'll
# be floating-point at times but we want them to be
# accurately-represented integers when converted to strings or the
# indexing will break. The code appears to work as-is, anyway.



BEGIN {

# Supporting old awk makes it effectively impossible to take a general
# approach which supports all terminals, it seems to me.
#
# So instead the game uses a few VT100 control codes. To use a generic
# approach, uncomment the next line (by removing the #).
#novt=1

# This should be 18 to never omit any cards from the display,
# but that requires a 26-line terminal (or removing the blank spacing
# lines, which looks a bit odd).
if(displines<=0) displines=16

# V7 seems to need this
if(novt=="") novt=0

if(novt)
  {
  # Generic approach, uglier but usable
  clearstr="\n\n\n"
  homestr="\n\n\n"
  cleolstr=""
  }
else
  {
  # VT100/ANSI/xterm etc.
  # (old awk doesn't handle octal escapes (?) so literal Esc, enjoy!)
  escstr=""
  clearstr=escstr "[H" escstr "[J"
  homestr=escstr "[H"
  cleolstr=escstr "[K"
  }

valuetbl="A23456789TJQK"
suittbl="HDsc"

printf("%s",clearstr)

moveprompt="Enter move, or ? for help > "
waitprompt="Press Enter to continue > "

print "zpsoli 1.0 - portable Solitaire/Patience/Klondike card game in awk\n\n"

# Supporting both our own RNG for old awk and the provided RNG in new
# awk is complicated by old awk freaking out if it spots srand() and
# rand(), which it will (given the implementation) even if they're not
# reached at runtime.
#
# The solution I've used is to have two "executables", one is this
# as-is which uses the new-awk RNG, and one (made via sed in the
# Makefile) sets the "nawkrng" flag to zero below and removes srand()
# and rand(). The latter version works on both new awk and old awk,
# using the old-awk approach either way.

# An initial prompt is necessary to avoid heavy code duplication,
# given that old awk lacks user functions. In old awk mode it's also
# used to prompt for a random-number seed.

# nawkrng must be zero *and srand/rand removed* for the old-awk version.
nawkrng=1

if(nawkrng)
  {
  srand()
  printf("Press Enter to start > ")
  }
else
  {
  printf("Please enter random-number seed (e.g. time as HHMMSS) > ")
  }

loweronly=0
waitline=1
dorestart=1
}


# The "NR>0" below is unnecessary, but the awk used in Minix 2 and
# Minix 3.1.[0-6] completely fails to run this action without it
# (breaking the game), possibly because we otherwise have two actions
# with the same pattern (i.e. none). It somehow runs out of memory
# before you can win though.
NR>0 {
inputok=0

# needed to avoid having to set $0, which old awk disallows
inlin=$0
}


waitline==1 {
# Used to cope with old awk not having getline.
# Any input is replaced with the "redraw" command :-)
# (but it's not actually used due to setting inputok).
inlin="redraw"
chstr=clearstr
inputok=1
waitline=0

if(NR==1 && !nawkrng)
  {
  rndseed=0+$0
  if(rndseed<0.) rndseed=0.
  rndseed%=65536.
  }
}


# new and restart must be typed in full
# (any remaining input on the line is ignored)
inlin ~ /^new/ || inlin ~ /^restart/ {
inputok=1
dorestart=1
}


dorestart==1 {
# clear all piles, generate new deck, and shuffle it

# zero out piles array
for(f=0;f<13;f++)
  {
  piles[f ":0"]=""
  plast[f]=-1
  }

# generate a new face-down deck in order
ndnum=0
for(ndsuit=0;ndsuit<4;ndsuit++)
  {
  for(ndrank=1;ndrank<=13;ndrank++)
    {
    newdeck[ndnum]=ndsuit "1" ndrank
    ndnum++
    }
  }

newdeck[52]=""

# now, to shuffle them, move a random card to the real deck, 52 times.
# technically this isn't a true shuffle, but we're more interested in
# randomness than replicating the physical act of shuffling IMHO. :-)

dstnum=0

for(f=52;f>=1;f--)
  {
  # get random number, mod f
  
  if(nawkrng)
    {
    rndnum=rand()
    }
  else
    {
    # The original soliyawn effectively used this algorithm -
    # a Lehmer RNG with m=65537 and a=75 as used by the Spectrum ROM,
    # see: https://en.wikipedia.org/wiki/Lehmer_random_number_generator
    # It's tweaked to avoid zero and get a 16-bit result, as
    # zero must not be used as the actual seed, and thus will never be
    # generated by the multiplication. Not the greatest algorithm by
    # modern standards, but tolerable enough for this purpose, and not
    # hard to do in awk.
    # ("int" here in case of cumulative FP precision issues.)
    rndseed=int(((rndseed+1)*75.)%65537.-1)
    
    # There's a BSD-derived old awk for DOS that doesn't like this as
    # "rndnum=rndseed/65536" (it sees the / as an incomplete regexp
    # somehow!), but it will tolerate this version.
    rndnum=rndseed
    rndnum/=65536.
    }
  
  # Since it's better to use the upper bits, the traditional Basic
  # approach of e.g. int(rnd*range) for 0..range-1 is appropriate.
  # Well, that's true for the RNG above at least. :-)
  srcnum=int(rndnum*f)
  piles[0 ":" dstnum]=newdeck[srcnum]
  dstnum++
  
  # Move the rest down to close the gap.
  # (This isn't a very efficient approach, to say the least, and
  # may result in a significant delay on 70s/80s-era machines.)
  while(newdeck[srcnum]!="")
    {
    newdeck[srcnum]=newdeck[srcnum+1]
    srcnum++
    }
  }

piles["0:52"]=""
plast[0]=51


# do the initial dealing of cards to the main piles.
lastcard=51
for(pnum=2;pnum<=8;pnum++)
  {
  for(cnum=0;cnum<pnum-1;cnum++)
    {
    piles[pnum ":" cnum]=piles["0:" lastcard]
    lastcard--
    # we don't update the deck end-of-pile entry until done
    }
  
  # last card is face-up
  cardstr=piles[pnum ":" cnum-1]
  piles[pnum ":" cnum-1]=substr(cardstr,1,1) "0" substr(cardstr,3,2)

  # add end-of-pile
  piles[pnum ":" cnum]=""
  plast[pnum]=cnum-1
  }

# update end-of-pile entry for deck
piles["0:" lastcard+1]=""
plast[0]=lastcard

chstr=clearstr

# restart (or init) finished now
dorestart=0
}


inlin ~ /^quit/ || inlin ~ /^exit/ {
print ""
exit
}


inlin ~ /^\?/ || inlin ~ /^h/ {
printf("%s",clearstr)
print "=== Move and Command Help ==="
print ""
print "Press Enter alone to draw a card from the deck."
print "To move cards, enter source then destination:"
print ""
print "d = last drawn card"
print "1-7 = piles 1-7"
print "a = ace etc. piles"
print ""
print "Examples:"
print ""
print "25   (move card or run from pile 2 to 5)"
print "da   (move last drawn card to ace piles)"
print ""
print "On piles 1-7 suit colours must alternate (indicated with uppercase"
print "vs. lowercase), on ace piles suit must match existing cards."
print "You win if the ace piles reach all-kings."
print ""
print "Aside from moves, the (self-explanatory) commands available are"
print "help, redraw, restart (or new), and quit (or exit)."
print "Use novt for a simpler display if output looks strange."
print ""
printf(waitprompt)

waitline=1
next
}


# Enter only = draw a card
# (This was previously "inlin ~ /^$/" which was stupid, but also broke
# awka for some reason.)
inlin=="" {
inputok=1

# turn over a card, i.e. move one from the deck to the turned-over pile.
# If the deck is empty when this is called, it moves the turned-over
# pile to the deck (in reverse order, to simulate turning the cards
# back over again), then turns over a card. Or, if no cards are left,
# it's just a nop.

# see if the deck is empty
toempty=0
if(plast[0]==-1)
  {
  # move turned-over pile to deck in reverse order...
  # start by finding last card in turned-over pile
  num=plast[1]

  if(num==-1)
    {
    # turned-over pile is empty
    toempty=1
    }
  else
    {
    # num is last card in turned-over pile
    dnum=0
    for(f=num;f>=0;f--)
      {
      # copy, making sure to put them face-down
      cardstr=piles["1:" f]
      piles["0:" dnum]=substr(cardstr,1,1) "1" substr(cardstr,3,2)
      dnum++
      }

    # update end-of-pile on deck and turned-over cards
    piles["0:" dnum]=""
    plast[0]=dnum-1
    piles["1:0"]=""
    plast[1]=-1
    }
  }

# no error for empty deck/turned-over, we just fall through instead.

if(!toempty)
  {
  # deck has one or more cards, turn top one over.
  srcnum=plast[0]
  dstnum=plast[1]
  
  # we pretend previously turned-over cards are face-down, to make
  # the card-move code elsewhere slightly easier.
  if(dstnum!=-1)
    {
    cardstr=piles["1:" dstnum]
    piles["1:" dstnum]=substr(cardstr,1,1) "1" substr(cardstr,3,2)
    }

  # now move card from deck, and put it face-up
  cardstr=piles["0:" srcnum]
  piles["0:" srcnum]=""
  plast[0]=srcnum-1
  dstnum++
  piles["1:" dstnum]=substr(cardstr,1,1) "0" substr(cardstr,3,2)
  piles["1:" dstnum+1]=""
  plast[1]=dstnum
  }

# set to only update lower display if possible
if(!novt && chstr==homestr)
  loweronly=1
}


# movement commands are all two chars, source and dest.
# possible locations are:
#
# d=deck (source only)
# 1..7=normal piles
# a=ace piles (dest only)
#
# We ignore any extra after first two chars if it matches.

inlin ~ /^[1-7dD][1-7aA]/ {
inputok=1

# convert d/D/a/A to 1, and 1..7 to 2..8
# aces pile number is corrected later as needed.
srcpile=substr(inlin,1,1)+1
dstpile=substr(inlin,2,1)+1

# do card(s) movement
if(dstpile==1)
  {
  # ace piles handled specially
  # find last card in source pile
  srcnum=plast[srcpile]
  if(srcnum!=-1)
    {
    # set correct ace pile based on card suit
    cardstr=piles[srcpile ":" srcnum]
    dstpile=9+substr(cardstr,1,1)

    # check existing card (we only keep the top one for ace piles).
    # the source card value must be the next one along from that on the
    # ace pile (this works even if the ace one is missing :-))
    srcrank=0+substr(cardstr,3,2)
    dstrank=1+substr(piles[dstpile ":0"],3,2)
    if(srcrank==dstrank)
      {
      # finally, if it's ok, move it there.
      piles[dstpile ":0"]=cardstr
      piles[srcpile ":" srcnum]=""
      plast[srcpile]=srcnum-1
      }
    }

  # end of move-to-ace-piles
  }
else
  {
  # move to normal pile
  #
  # so srcpile is src pile, dstpile is dest pile. The cards we
  # compare to check this is ok are the one at the
  # base of the src pile's run (the lowest one which
  # is face-up), and the lowest card on the dest pile.
  # This even works when the turned-over pile is src,
  # due to a nice little kludge in the turnover code. :-)

  if(srcpile!=dstpile && piles[srcpile ":0"]!="")
    {
    # so the piles differ, and source is non-empty.
    # find first face-up card, going from bottom of src pile.
    srcnum=-1
    for(f=0;srcnum<0 && f<60;f++)
      if(substr(piles[srcpile ":" f],2,1)=="0")
	srcnum=f
    
    # now get last card on dest pile
    dstnum=plast[dstpile]

    # see if move is ok to do
    moveok=0
    srcstr=piles[srcpile ":" srcnum]
    srcrank=substr(srcstr,3,2)
    
    if(dstnum<0)
      {
      # if dest is empty, top of src run must be a king.
      if(srcrank=="13") moveok=1
      }
    else
      {
      # if dest not empty, dest card must be one higher,
      # and must have opposite suit colour.
      dststr=piles[dstpile ":" dstnum]
      srccol=0
      dstcol=0
      if(substr(srcstr,1,1)>=2) srccol=1
      if(substr(dststr,1,1)>=2) dstcol=1
      if(substr(srcstr,3,2)+1==substr(dststr,3,2)+0 && srccol+dstcol==1)
        moveok=1
      }

    if(moveok)
      {
      # if we got here, it's ok to do the move.
      # copy all cards to the end, updating end-of-pile markers after.
      origsrcnum=srcnum
      dstnum++
      while(piles[srcpile ":" srcnum]!="")
        {
	piles[dstpile ":" dstnum]=piles[srcpile ":" srcnum]
	srcnum++
	dstnum++
	}

      piles[srcpile ":" origsrcnum]=""
      plast[srcpile]=origsrcnum-1
      piles[dstpile ":" dstnum]=""
      plast[dstpile]=dstnum-1
      }
    }
  }
}


inlin ~ /^novt/ {
inputok=1
novt=1
clearstr="\n\n\n"
homestr=clearstr
cleolstr=""
chstr=clearstr
}


# redraw
inputok==0 && inlin ~ /^r/ {
inputok=1
chstr=clearstr
}


# reaching here without something sensible gives an error
inputok==0 {
print ""
print "Move/command not understood - try ? for help"
print ""

# skip display for this one, and clear next time
chstr=clearstr
printf(cleolstr moveprompt)
next
}


{
# after responding to the move/command input above,
# we now display piles, and show prompt for next move

# but first, if any cards at the end of main piles are face down
# after a move, turn them up.
# (turned-over cards (pile 1) are included too, due to face-down hack)
for(pile=1;pile<=8;pile++)
  {
  num=plast[pile]
  if(num>=0)
    {
    cardstr=piles[pile ":" num]
    piles[pile ":" num]=substr(cardstr,1,1) "0" substr(cardstr,3,2)
    }
  }

# do display
# first the main piles
printf("%s",chstr)
chstr=homestr

if(loweronly)
  {
  # skip the upper display (if possible) when only changing the deck -
  # could move cursor directly, but this keeps it to only having to
  # specify clear/home/cleolstr in BEGIN.
  loweronly=0
  for(y=0;y<displines+2;y++)
    print ""
  }
else
  {
  # do upper display

  for(pile=2;pile<=8;pile++)
    {
    empty[pile]=0
    if(piles[pile ":0"]=="") empty[pile]=1
    done[pile]=0
    doneby[pile]=0
    }
  
  print "   1   2   3   4   5   6   7"
  print ""
  
  # displines should really be 18 to allow visibility for all cases, but
  # the default is 16 so that the display fits in 24 lines (without
  # having to interleave the upper display with the lower one).
  
  for(y=0;y<displines;y++)
    {
    for(pile=2;pile<=8;pile++)
      {
      cardstr=piles[pile ":" y]
      if(done[pile] || cardstr=="")
        {
        done[pile]=1
        if(doneby[pile]<2 && !empty[pile])
  	printf("  ..")
        else
  	printf("    ")
        doneby[pile]++
        }
      else
        {
        if(substr(cardstr,2,1)!="0")
    	  printf("  %%%%")
        else
          {
	  printf("  %s%s",\
	         substr(valuetbl,substr(cardstr,3,2),1),\
	         substr(suittbl,1+substr(cardstr,1,1),1))
	  }
	}
      }
    
    print ""
    }
  }

# now do deck (pile 0), turned-over (1), a gap, then ace piles (9-12)
for(y=0;y<3;y++)
  {
  for(pile=0;pile<7;pile++)
    {
    if(pile==2)
      {
      # mind the gap
      printf("    ")
      }
    else
      {
      if(pile<2)
        {
        # these are proper piles, so we need to find the last card
	if(y==0)
	  {
          num=plast[pile]
          cardstr=""
          if(num>=0) cardstr=piles[pile ":" num]
          oldstr[pile]=cardstr
          }
        else
          cardstr=oldstr[pile]
        }
      else
        {
        # for aces we just use the only card
        cardstr=piles[pile+6 ":0"]
        }

      # display card
      if(cardstr=="")
        printf("    ")
      else
        {
        if(y==0)
          {
          if(substr(cardstr,2,1)!="0")
            printf("%%%%  ")
          else
            {
            printf("%s%s  ",\
                   substr(valuetbl,substr(cardstr,3,2),1),\
                   substr(suittbl,1+substr(cardstr,1,1),1))
            }
          }
        else
          {
          if(substr(cardstr,2,1)!="0")
            printf("%%%%  ")
          else
            printf("..  ")
          }
        }
      }
    }

  print ""
  }


# see if they've won - all the ace slots will contain kings if so.
gamewon=1
for(f=9;f<=12;f++)
  {
  if(substr(piles[f ":0"],3,2)!="13")
    gamewon=0
  }

if(gamewon)
  {
  print ""
  print "=== Congratulations - you won! ==="
  print ""
  printf(waitprompt)

  waitline=1
  dorestart=1
  next
  }

print ""
printf(cleolstr moveprompt)
}
